import codecs
import enum
import re

from scruf import compare, exception, test


class TestParser:
    """
    Parser for parsing raw tests

    Attributes
    ----------
    indent : string
        The indentation to be used when parsing tests
    """

    def __init__(self, indent="    "):
        self.tokeniser = _TestTokeniser(indent)
        self.fsm = _TestFSM(self.tokeniser)

        self._content_getter = {
            self.fsm.States.DESCRIPTION: self.tokeniser.get_description,
            self.fsm.States.SETUP_COMMAND: self.tokeniser.get_setup_command,
            self.fsm.States.TEST_COMMAND: self.tokeniser.get_test_command,
            self.fsm.States.CONTINUE: self.tokeniser.get_continue,
            self.fsm.States.OUTPUT_SPEC: self.tokeniser.get_result,
        }

    def parse(self, lines):
        """Parse raw lines into `scruf.test.Test` objects

        Parameters
        ----------
        lines : list of str

        Returns
        -------
        list of `scruf.test.Test`

        Raises
        ------
        scruf.parse.ProgressionError
            If an invalid progression is found, e.g. the raw lines hold a description
            followed by a continuation line
        """

        prev_state = self.fsm.States.START
        tests = []
        test = self._new_test()

        for i, line in enumerate(lines):
            if self.tokeniser.is_comment(line) or self.tokeniser.is_empty(line):
                continue

            state = self.fsm.progress(prev_state, line)
            if state == self.fsm.States.START:
                tests.append(self._test_to_test_object(test))
                test = self._new_test()
                state = self.fsm.progress(state, line)

            if state is None:
                from_state = self.fsm.state_name(prev_state)
                expected_states = self.fsm.get_named_progressions(prev_state)
                raise ProgressionError(i, line, from_state, expected_states)
            # Preserve newlines for results
            if not state == self.fsm.States.OUTPUT_SPEC:
                line = line.rstrip()

            test[state].append(self._content_getter[state](line))
            prev_state = state

        tests.append(self._test_to_test_object(test))
        return tests

    def _new_test(self):
        return dict((key, []) for key in self._content_getter)

    def _test_to_test_object(self, raw_test):
        command = " ".join(
            raw_test[self.fsm.States.TEST_COMMAND] + raw_test[self.fsm.States.CONTINUE]
        )
        description = " ".join(raw_test[self.fsm.States.DESCRIPTION])
        output_specs = raw_test[self.fsm.States.OUTPUT_SPEC]
        setup_commands = raw_test[self.fsm.States.SETUP_COMMAND]
        return test.Test(
            command,
            setup_commands=setup_commands,
            description=description,
            output_specs=output_specs,
        )


class OutputParser:
    _EOL_CHARACTERS = "\n\r"
    _STDOUT_FLAG = "1:"
    _STDERR_FLAG = "2:"
    _REGEX_FLAG = "[RE]"
    _NO_EOL_FLAG = "[NEoL]"
    _ESCAPE_FLAG = "[ESC]"
    _RETURNCODE_REGEX = re.compile(r"\[(\d+)\]\s*$")

    @classmethod
    def parse(cls, raw_spec):
        """Parse a raw output spec line and extract details

        Parameters
        ----------
        raw_spec : str
            The test line to process

        Returns
        -------
        parse.test.OutputSpec
        """
        content, source = cls._get_source_and_content(raw_spec)

        if cls._startswith_flag(content, cls._REGEX_FLAG):
            content = cls._strip_flag(content, cls._REGEX_FLAG).rstrip(
                cls._EOL_CHARACTERS
            )
            return test.OutputSpec(
                comp_type=compare.ComparisonTypes.REGEX, content=content, source=source,
            )
        if cls._startswith_flag(content, cls._NO_EOL_FLAG):
            content = cls._strip_flag(content, cls._NO_EOL_FLAG).rstrip(
                cls._EOL_CHARACTERS
            )
            return test.OutputSpec(
                comp_type=compare.ComparisonTypes.NO_EOL,
                content=content,
                source=source,
            )
        if cls._startswith_flag(content, cls._ESCAPE_FLAG):
            content = cls._get_escape_content(
                cls._strip_flag(content, cls._ESCAPE_FLAG)
            )
            return test.OutputSpec(
                comp_type=compare.ComparisonTypes.ESCAPE,
                content=content,
                source=source,
            )
        if cls._has_returncode_flag(content):
            content = cls._get_returncode(content)
            return test.OutputSpec(
                comp_type=compare.ComparisonTypes.RETURNCODE,
                content=content,
                source="returncode",
            )
        return test.OutputSpec(
            comp_type=compare.ComparisonTypes.BASIC, content=content, source=source,
        )

    @classmethod
    def _get_source_and_content(cls, line):
        if cls._startswith_flag(line, cls._STDOUT_FLAG):
            return cls._strip_flag(line, cls._STDOUT_FLAG), "stdout"
        if cls._startswith_flag(line, cls._STDERR_FLAG):
            return cls._strip_flag(line, cls._STDERR_FLAG), "stderr"
        return line, "stdout"

    @classmethod
    def _has_returncode_flag(cls, line):
        return cls._RETURNCODE_REGEX.match(line) is not None

    @classmethod
    def _get_returncode(cls, line):
        return int(cls._RETURNCODE_REGEX.match(line).group(1))

    @staticmethod
    def _get_escape_content(line):
        return codecs.decode(line, "unicode-escape")

    @staticmethod
    def _startswith_flag(line, flag):
        return line.startswith(flag)

    @staticmethod
    def _strip_flag(line, flag):
        return OutputParser._lstrip_space(line[len(flag) :])

    @staticmethod
    def _lstrip_space(line):
        if line.startswith(" "):
            return line[1:]
        return line


class ProgressionError(exception.CramerError):
    """
    Exception for invalid progressions when parsing raw lines

    Attributes
    ----------
    line_num : int
    line_content : str
    from_state : str
        Human readable version of the state before the invalid line
    expected_states : list of str
        List of human readable version of the expected states given `from_state`

    """

    def __init__(self, line_num, line_content, from_state, expected_states):
        message = "Failed to parse line {}: {}\nExpected line of type: {}".format(
            line_num, line_content.rstrip(), " or ".join(expected_states),
        )
        super().__init__(message)

        self.line_num = line_num
        self.from_state = from_state
        self.expected_states = expected_states


class _TestTokeniser:
    def __init__(self, indent):
        self.comment_character = "#"
        self.command_character = "$"
        self.continue_character = ">"
        self.indent = indent

        self.command_prefix = self.indent + self.command_character
        self.continue_prefix = self.indent + self.continue_character

    def is_test_command(self, line):
        return line.startswith(self.command_prefix)

    def is_setup_command(self, line):
        return line.startswith(self.command_character)

    def is_continue(self, line):
        return line.startswith(self.continue_prefix)

    def is_description(self, line):
        return (
            not line[0].isspace()
            and not self.is_comment(line)
            and not self.is_setup_command(line)
        )

    def is_comment(self, line):
        return line.startswith(self.comment_character)

    def is_empty(self, line):
        return not line or not line.startswith(self.indent) and line.isspace()

    def is_result(self, line):
        return (
            line.startswith(self.indent)
            and not self.is_test_command(line)
            and not self.is_continue(line)
        )

    def get_test_command(self, line):
        return line[len(self.command_prefix) :].strip()

    def get_setup_command(self, line):
        return line[len(self.command_character) :].strip()

    def get_continue(self, line):
        return line[len(self.continue_prefix) :].strip()

    def get_description(self, line):
        return line

    def get_result(self, line):
        return line[len(self.indent) :]


class _TestFSM:
    class States(enum.Enum):
        START = enum.auto()
        DESCRIPTION = enum.auto()
        SETUP_COMMAND = enum.auto()
        TEST_COMMAND = enum.auto()
        CONTINUE = enum.auto()
        OUTPUT_SPEC = enum.auto()

    STATE_NAMES = {
        States.START: "start",
        States.DESCRIPTION: "description",
        States.SETUP_COMMAND: "setup_command",
        States.TEST_COMMAND: "test_command",
        States.CONTINUE: "continue",
        States.OUTPUT_SPEC: "result",
    }

    def __init__(self, tokeniser):
        self.tokeniser = tokeniser

        self.progressions = {
            self.States.START: {
                self.States.DESCRIPTION: self.tokeniser.is_description,
                self.States.SETUP_COMMAND: self.tokeniser.is_setup_command,
                self.States.TEST_COMMAND: self.tokeniser.is_test_command,
            },
            self.States.DESCRIPTION: {
                self.States.DESCRIPTION: self.tokeniser.is_description,
                self.States.SETUP_COMMAND: self.tokeniser.is_setup_command,
                self.States.TEST_COMMAND: self.tokeniser.is_test_command,
            },
            self.States.SETUP_COMMAND: {
                self.States.SETUP_COMMAND: self.tokeniser.is_setup_command,
                self.States.TEST_COMMAND: self.tokeniser.is_test_command,
            },
            self.States.TEST_COMMAND: {
                self.States.CONTINUE: self.tokeniser.is_continue,
                self.States.OUTPUT_SPEC: self.tokeniser.is_result,
                self.States.START: lambda line: self.tokeniser.is_description(line)
                or self.tokeniser.is_test_command(line)
                or self.tokeniser.is_setup_command(line),
            },
            self.States.CONTINUE: {
                self.States.CONTINUE: self.tokeniser.is_continue,
                self.States.OUTPUT_SPEC: self.tokeniser.is_result,
                self.States.START: lambda line: self.tokeniser.is_description(line)
                or self.tokeniser.is_test_command(line),
            },
            self.States.OUTPUT_SPEC: {
                self.States.OUTPUT_SPEC: self.tokeniser.is_result,
                self.States.START: lambda line: self.tokeniser.is_description(line)
                or self.tokeniser.is_test_command(line),
            },
        }

    def progress(self, from_state, line):
        for to_state, condition in self.progressions[from_state].items():
            if condition(line):
                return to_state
        return None

    @classmethod
    def state_name(cls, state):
        return cls.STATE_NAMES[state]

    def get_named_progressions(self, state):
        progression_states = list(self.progressions[state].keys())

        # "Start" is just an internal state, so avoid presenting it
        if self.States.START in progression_states:
            del progression_states[progression_states.index(self.States.START)]
            progression_states += self.progressions[self.States.START].keys()
        return [self.state_name(s) for s in progression_states]



"""Test runner"""
from scruf import compare, parse


def compare_result(test, exec_result):
    """Compare a test's expected output against an execute result

    Parameters
    ----------
    test : scruf.test.Test
    exec_result : scruf.execute.Result

    Returns
    -------
    list of dict
        a list of the test result for each test line. Each dict contains keys:
            "result_line": str
                The line from the `exec_result` that was compared
            "result_source": str
                The source of the `"result_line"`, one of "stdout" or "stderr"
            "test_line": str
                The line from the `test` that was compared
            "comparison_result": bool
            "comparison_type": scruf.compare.ComparisonTypes
                The type of comparison made

    Raises
    ------
    scruf.compare.RegexError
        If the test contains a regex comparison with an invalid regex
    scruf.execute.OutOfLinesError
        If there are more lines in the test than the `exec_result`
    """
    summary = []

    for line in test.output_specs:
        output_spec = parse.OutputParser.parse(line)

        result_content = exec_result.get_content_for_comparison(
            output_spec.source, output_spec.comp_type
        )
        comparison = output_spec.compare(result_content)
        summary.append(
            {
                "result_line": result_content,
                "result_source": output_spec.source,
                "test_line": output_spec.content,
                "comparison_result": comparison,
                "comparison_type": output_spec.comp_type,
            }
        )
    return summary


def get_printable_failures(comparison):
    """Get a printable representation of a failed comparison

    Parameters
    ----------
    comparison : dict
        As returned `compare_result`

    Returns
    -------
    str
    """

    comparison_type = comparison["comparison_type"]

    return {
        compare.ComparisonTypes.BASIC: _basic_comparison_failure_lines,
        compare.ComparisonTypes.ESCAPE: _basic_comparison_failure_lines,
        compare.ComparisonTypes.NO_EOL: _no_eol_comparison_failure_lines,
        compare.ComparisonTypes.REGEX: _regex_comparison_failure_lines,
        compare.ComparisonTypes.RETURNCODE: _returncode_comparison_failure_lines,
    }[comparison_type](comparison["test_line"], comparison["result_line"])


def _format_content(content):
    formatted_content = repr(content)

    if not content.endswith("\n"):
        formatted_content += " (no eol)"
    return formatted_content


def _basic_comparison_failure_lines(comparison_content, result_content):
    return [
        "got: " + _format_content(result_content),
        "expected: " + _format_content(comparison_content),
    ]


def _no_eol_comparison_failure_lines(comparison_content, result_content):
    return _basic_comparison_failure_lines(
        comparison_content.rstrip("\r\n"), result_content
    )


def _regex_comparison_failure_lines(comparison_content, result_content):
    return [
        "got: " + _format_content(result_content),
        "Does not match: '{}'".format(comparison_content),
    ]


def _returncode_comparison_failure_lines(expected_returncode, got_returncode):
    return [
        "got return code: {} != {}".format(
            str(got_returncode), str(expected_returncode)
        )
    ]



import sys

from scruf import observe, run


class TapObserver(observe.Observer):
    def notify_before_testing_file(self, filename):
        print(self._tap_comment_line("Testing: {}".format(filename)))

    def notify_before_tests_run(self, tests):
        print("1..{}".format(len(tests)))

    def notify_test_error(self, test, test_number, error):
        print(self._tap_failure(self._build_output_line(test, test_number)))
        print(self._tap_error_line(error), file=sys.stderr)

    def notify_test_success(self, test, test_number):
        print(self._tap_success(self._build_output_line(test, test_number)))

    def notify_test_comparison_failure(self, test, test_number, failed_comparisons):
        print(self._tap_failure(self._build_output_line(test, test_number)))
        for comparison in failed_comparisons:
            failure_lines = [
                self._tap_comment_line("\t" + line)
                for line in run.get_printable_failures(comparison)
            ]
            print("\n".join(failure_lines), file=sys.stderr)

    def notify_test_strict_failure(self, test, test_number, source_map):
        print(self._tap_failure(self._build_output_line(test, test_number)))
        for source, remaining_lines in source_map.items():
            description = "Content still remaining for {}:".format(source)
            print(self._tap_comment_line(description), file=sys.stderr)
            for line in remaining_lines:
                print(self._tap_comment_line("\t" + line.rstrip()), file=sys.stderr)

    ######################
    # TAP specific methods
    ######################
    @staticmethod
    def _tap_success(content):
        return "ok " + content

    @staticmethod
    def _tap_failure(content):
        return "not ok " + content

    @staticmethod
    def _tap_comment_line(content):
        return "# " + content

    @classmethod
    def _tap_error_line(cls, error):
        error_lines = []
        for line in str(error).splitlines():
            error_lines.append(cls._tap_comment_line("\t" + line))
        return "\n".join(error_lines)

    @staticmethod
    def _build_output_line(test, test_number):
        output_line = str(test_number)
        if test.description:
            output_line += " - {}".format(test.description.rstrip())
        return output_line



import os
import shutil
import subprocess
import tempfile

from scruf import compare, exception


class Executor:
    """
    Class for executing commands

    Attributes
    ----------
    shell : str
        The shell to be used to run the command
    cleanup : bool
        Weather or not to remove created temporary directories once a command is run
    path : str
        Path to use in the 'TESTDIR' environment variable when executing commands
    tmpdir : str
        Directory under which the testdir will be created for running commands
    env : dict
        Environment to be used when running the command, defaults to `os.environ`

    Raises
    ------
    scruf.execute.FailedToCreateTestDirError
        If unable to create a directory under `tmpdir`. Specifically, if attempting to
        create this directory raises `PermissionError`or `FileNotFoundError`
    """

    def __init__(
        self, shell="/bin/sh", cleanup=True, path=os.getcwd(), tmpdir="/tmp", env=None,
    ):
        self.cleanup = cleanup
        self.shell = shell
        self.env = self._setup_env(env, path)

        # Might raise exception, so set this last
        self.testdir = self._mktestdir(tmpdir)

    def execute(self, command):
        """Execute a command

        Parameters
        ----------
        command : str


        Returns
        -------
        tuple
            int : The return code from executing the command
            string : The stdout generated from the command
            string : The stderr generated from the command
        """

        p = subprocess.Popen(
            [self.shell, "-"],
            cwd=self.testdir,
            env=self.env,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            encoding="utf-8",
        )
        stdout, stderr = p.communicate(command)

        return Result(p.returncode, stdout, stderr)

    def __del__(self):
        if self.cleanup and hasattr(self, "testdir"):
            shutil.rmtree(self.testdir)

    @staticmethod
    def _setup_env(env, path):
        if env is None:
            env = os.environ
        _maybe_add_to_env(env, "TESTDIR", path)
        return env

    @staticmethod
    def _mktestdir(tmpdir):
        try:
            testdir = tempfile.mkdtemp(dir=tmpdir, suffix="scruf")
        except (PermissionError, FileNotFoundError) as e:
            raise FailedToCreateTestDirError(tmpdir, e)
        return testdir


class Result:
    """
    Represents the result of executing a command

    Attributes
    ----------
    returncode : int
    stdout : str
    stderr : str
    """

    def __init__(self, returncode, stdout="", stderr=""):
        self.stdout = stdout
        self.stderr = stderr
        self.returncode = returncode

        self._stdout_index = 0
        self._stdout_lines = stdout.splitlines(True)

        self._stderr_index = 0
        self._stderr_lines = stderr.splitlines(True)

    """Get the next line from the given source

    Parameters
    ----------
    source : { "stdout", "stderr" }
        The source to extra the next line from

    Returns
    -------
    str
        The next line from the source

    Raises
    ------
    scruf.execute.OutOfLinesError
        If there are no more lines available for the given source
    """

    def next_line(self, source):
        index_attr = self._index_for_source(source)
        index = getattr(self, index_attr)
        lines = getattr(self, self._lines_for_source(source))

        if index == len(lines):
            raise OutOfLinesError(self, source)
        else:
            setattr(self, index_attr, index + 1)
            line = lines[index]
            return line

    def has_remaining_lines(self, source):
        lines = getattr(self, self._lines_for_source(source))
        index = getattr(self, self._index_for_source(source))
        return index <= len(lines) - 1

    def get_remaining_lines(self, source):
        lines = getattr(self, self._lines_for_source(source))
        index = getattr(self, self._index_for_source(source))
        return lines[index:]

    def get_content_for_comparison(self, source, comp_type):
        if comp_type == compare.ComparisonTypes.RETURNCODE:
            return self.returncode
        return self.next_line(source)

    @staticmethod
    def _index_for_source(source):
        return "_" + source + "_index"

    @staticmethod
    def _lines_for_source(source):
        return "_" + source + "_lines"


class OutOfLinesError(exception.CramerError):
    def __init__(self, result, source):
        message = "No more lines available for {}:\n".format(source)
        message += "\t" + self._build_stream_content_msg(result, "stdout")
        message += "\t" + self._build_stream_content_msg(result, "stderr")
        super().__init__(message)

        self.result = result
        self.source = source

    @staticmethod
    def _build_stream_content_msg(result, stream):
        content = getattr(result, stream)
        msg = "{} was: ".format(stream)
        if content:
            return msg + content
        return msg + "(no content)\n"


class FailedToCreateTestDirError(exception.CramerError):
    def __init__(self, dir_name, file_error):
        message = "Could not create temporary directory at {}: {}".format(
            dir_name, str(file_error)
        )
        super().__init__(message)

        self.dir_name = dir_name


def _maybe_add_to_env(env, name, value):
    if name not in env:
        env[name] = value



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    author="Matthew Hughes",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Testing",
    ],
    description="Functional testing for command line applications",
    license="GNU GPLv3",
    long_description=long_description,
    name="scruf",
    packages=["scruf", "scruf.observers"],
    python_requires=">=3.6",
    scripts=["bin/scruf"],
    url="https://gitlab.com/matthewhughes/scruf",
    version="0.4",
)



"""Main entry point for CLI application"""
import argparse
import configparser
import re
import sys

import scruf
from scruf import compare, exception, execute, parse, run as scruf_run
from scruf.observers import tap


def run(argv):
    """Run scruf with options over some files

    Returns
    -------
    int
        0 if all tests pass, otherwise 1, suitable for sys.exit
    """
    options = _get_options(argv[1:])
    filenames = options.pop("files")
    try:
        _process_options(options)
    except OptionProcessingError as e:
        print("Failed to process options: {}".format(str(e)), file=sys.stderr)
        return 1

    observer = _build_observer()

    exit_code = 0 if run_files(filenames, options, observer) else 1
    return exit_code


def run_files(filenames, options, observer):
    success = True
    for filename in filenames:
        success &= run_file(filename, options, observer)
    return success


def run_file(filename, options, observer):
    try:
        with open(filename, "r") as f:
            lines = f.readlines()
    except (OSError, IOError) as file_error:
        observer.notify_file_open_error(filename, file_error)
        return False
    observer.notify_before_testing_file(filename)

    parser = parse.TestParser(options["indent"])
    try:
        tests = parser.parse(lines)
    except parse.ProgressionError as e:
        observer.notify_test_progression_error(filename, e)
        return False

    try:
        executor = execute.Executor(
            shell=options["shell"], cleanup=options["cleanup"], env=options["env"]
        )
    except execute.FailedToCreateTestDirError as e:
        observer.notify_execute_setup_error(filename, e)
        return False

    return run_tests(tests, executor, observer, options["strict"])


def run_tests(tests, executor, observer=None, strict=False):
    success = True
    observer.notify_before_tests_run(tests)

    for i, test in enumerate(tests):
        test_number = i + 1
        observer.notify_before_test_run(test, test_number)

        if not _run_setup_commands(test, test_number, executor, observer):
            success = False
            continue

        exec_result = executor.execute(test.command)
        try:
            compare_output = scruf_run.compare_result(test, exec_result)
        except (compare.RegexError, execute.OutOfLinesError) as e:
            observer.notify_test_error(test, test_number, e)
            success = False
            continue

        if strict:
            if not _check_for_strict_failure(test, test_number, exec_result, observer):
                success = False
                continue

        failure_summaries = [r for r in compare_output if not r["comparison_result"]]
        if len(failure_summaries) == 0:
            observer.notify_test_success(test, test_number)
        else:
            observer.notify_test_comparison_failure(
                test, test_number, failure_summaries
            )
            success = False
    return success


def _check_for_strict_failure(test, test_number, result, observer):
    strict_success = True
    observer_data = {}
    for source in ("stdout", "stderr"):
        if result.has_remaining_lines(source):
            observer_data[source] = result.get_remaining_lines(source)
            strict_success = False

    if not strict_success:
        observer.notify_test_strict_failure(test, test_number, observer_data)
    return strict_success


def _run_setup_commands(test, test_number, executor, observer):
    for setup_command in test.setup_commands:
        setup_result = executor.execute(setup_command)

        if setup_result.returncode != 0:
            observer.notify_setup_command_failure(
                test, test_number, setup_command, setup_result
            )
            return False
    return True


def _get_options(args):
    parser = argparse.ArgumentParser()
    parser.add_argument("files", metavar="FILE", nargs="+", help="File(s) to be tested")
    parser.add_argument(
        "--no-cleanup",
        action="store_false",
        dest="cleanup",
        help="Avoid cleaning up temporary test directory",
    )
    parser.add_argument(
        "-s",
        "--shell",
        type=str,
        help="Path to shell to be used to run tests with. Default is '/bin/sh'",
        default="/bin/sh",
    )
    parser.add_argument(
        "-i",
        "--indent",
        type=_indent_arg_type,
        default="    ",
        help="String to be used for detecting indentation when parsing tests. Default \
            is 4 spaces, use a literal '\\t' to denote a tab character",
    )
    parser.add_argument(
        "-e",
        "--env-file",
        type=str,
        help="Name of config file to read environment variables from",
    )
    parser.add_argument(
        "--version", action="version", version=str(scruf.__version__),
    )
    parser.add_argument(
        "--strict",
        action="store_true",
        help="Whether tests should be run in strict mode. In this mode a test that \
            does not check each line of output is considered to have failed",
    )

    return vars(parser.parse_args(args))


def _process_options(options):
    if options["env_file"] is not None:
        env = _get_env_from_file(options["env_file"])
        options["env"] = env
    else:
        options["env"] = None
    options.pop("env_file")


def _indent_arg_type(string):
    # Convert literal "\t" to tab characters, e.g. for "--indent '\t'"
    string = re.sub(r"\\t", "\t", string)

    if not string.isspace():
        msg = "{} is not entirely whitespace".format(string)
        raise argparse.ArgumentTypeError(msg)
    return string


def _build_observer():
    # TODO: once more than one observer exists this should take an option to decide
    # which to use
    return tap.TapObserver()


def _get_env_from_file(filename):
    scruf_env_key = "Scruf Env"
    config = configparser.ConfigParser()
    # make config case-sensitive
    config.optionxform = str
    try:
        config.read(filename)
    except configparser.ParsingError as e:
        raise OptionProcessingError(
            "Unable to parse config file '{}': {}".format(filename, str(e)),
        )

    if scruf_env_key not in config:
        raise OptionProcessingError(
            "Could not find expected section: '{}' in config file: {}".format(
                scruf_env_key, filename
            ),
        )

    return config[scruf_env_key]


class OptionProcessingError(exception.CramerError):
    def __init__(self, msg=None):
        super().__init__(msg)



"""Comparison logic for tests"""
import enum
import re

from scruf import exception

_EOL_CHARACTERS = "\n\r"


class ComparisonTypes(enum.Enum):
    """The possible types of comparison"""

    BASIC = enum.auto()
    ESCAPE = enum.auto()
    REGEX = enum.auto()
    NO_EOL = enum.auto()
    RETURNCODE = enum.auto()


def get_comparer(comparison_type):
    """Get function to be used to make comparison for the given type

    Note if `comparison_type` is `scruf.compare.ComparisonTypes.REGEX`, the returned
    function will raise a `scruf.compare.RegexError` if it is run with an invalid regex.

    Parameters
    ----------
    comparison_type : scruf.compare.ComparisonTypes

    Returns
    -------
    function(expected_content, exec_result_content)
        Function that can be used to make the comparison, returning bool
    """
    return {
        ComparisonTypes.BASIC: _basic_compare,
        ComparisonTypes.ESCAPE: _basic_compare,
        ComparisonTypes.REGEX: _regex_compare,
        ComparisonTypes.NO_EOL: _no_eol_compare,
        ComparisonTypes.RETURNCODE: _returncode_compare,
    }[comparison_type]


class RegexError(exception.CramerError):
    def __init__(self, regex, regex_error):
        message = "Error in regex {}: {}".format(regex, str(regex_error))
        super().__init__(message)

        self.regex = regex


def _basic_compare(expected_content, content):
    return content == expected_content


def _returncode_compare(expected_returncode, returncode):
    return returncode == expected_returncode


def _no_eol_compare(expected_content, content):
    return _basic_compare(expected_content.rstrip(_EOL_CHARACTERS), content)


def _build_regex(test_line):
    try:
        regex = re.compile(test_line)
    except re.error as e:
        raise RegexError(test_line, e)
    return regex


def _regex_compare(raw_regex, content):
    regex = _build_regex(raw_regex)
    return regex.search(content) is not None



import sys


class Observer:
    """
    Abstract class for observing test runs
    """

    def notify_file_open_error(self, filename, exception):
        """Notify when there is an error opening a test file

        Default implementation: print error to stderr

        Dispatched when a call to `open` on a test file raises either `OSError` or
        `IOError`. At this point we bail on testing the file

        Parameters
        ----------
        filename : str
            The file that opening was attempted on
        exception : { "OSError", "IOError" }
        """
        print(
            "Failed to open file {}: {}".format(filename, exception), file=sys.stderr,
        )

    def notify_test_progression_error(self, filename, exception):
        """Notify when there is an error trying to parse a file

        Default implementation: print error to stderr

        Dispatched when a call to `scruf.parse.TestParser.parse` raises a
        `scruf.parser.ProgressionError`. At this point we bail on testing the file

        Parameters
        ----------
        filename : str
        exceptoin : scruf.parse.ProgressionError
        """
        print(
            "Error when parsing {}: {}".format(filename, exception), file=sys.stderr,
        )

    def notify_execute_setup_error(self, filename, exception):
        """Notify when there is an error constructing an executor object for a test file

        Default implementation: print error to stderr

        Dispatched when a call to `scruf.execute.Executor` raises a
        `execute.FailedToCreateTestDirError` exception. At this point we bail on testing
        the file

        Parameters
        ----------
        filename : str
        exception : scruf.execute.FailedToCreateTestDirError
        """
        print(
            "Failed to setup test environment for {}: {}".format(filename, exception),
            file=sys.stderr,
        )

    def notify_before_testing_file(self, filename):
        """Notify once we've opened a file for testing

        Default implementation: noop

        Parameters
        ----------
        filename : str
        """
        pass

    def notify_before_tests_run(self, tests):
        """Notify before a set of tests are run

        Default implementation: noop

        Parameters
        ----------
        tests : list of `scruf.test.Test`
            The tests to be run
        """
        pass

    def notify_setup_command_failure(self, test, test_number, setup_command, result):
        """Notify when a setup command is run and returns nonzero

        Default implementation: noop

        If this is triggered we will bail on the current test

        Parameters
        ----------
        test : scurf.test.Test
            The test the setup command belongs to
        setup_command : str
        result : scruf.execute.Result
        """

        pass

    def notify_before_test_run(self, test, test_number):
        """Notify before a single test is run

        Default implementation: noop

        Parameters
        ----------
        test : scruf.test.Test
            The test to be run

        test_number : int
            1-based index of the test in the tests being run
        """
        pass

    def notify_test_error(self, test, test_number, error):
        """Notify when there is a failure running a test

        Default implementation: noop

        This will be called if an error arises from `scruf.run.compare_result()`. If
        this occurs the test will be skipped.

        Parameters
        ----------
        test : scruf.test.Test
        test_number : int
        error : { "scruf.compare.RegexError", "scruf.execute.OutOfLinesError" }
            The error that occurred during `scruf.run.compare_result()`

        """
        pass

    def notify_test_success(self, test, test_number):
        """
        Notify when a test succeeded

        Default implementation: noop

        Parameters
        ----------
        test : scruf.test.Test
        test_number : int
        """
        pass

    def notify_test_comparison_failure(self, test, test_number, failed_comparisons):
        """Notify when a test fails

        Default implementation: noop

        Parameters
        ----------
        test : scruf.test.Test
        test_number : int
        failed_comparisons : list of dicts
            The list of failed comparisons, i.e. outputs from `scruf.run.compare_result`
            with "comparison_result" == False
        """
        pass

    def notify_test_strict_failure(self, test, test_number, source_map):
        """Notify when a test fails a strict condition

        Default implementation: noop

        Parameters
        ----------
        test : scruf.test.Test
        test_number : int
        source_map : dict of strings to list of strings
            A map of the sources on output lines and their content lines
        """
        pass



from scruf import compare


class Test:
    """
    A runable test

    Attributes
    ----------
    command : string
        The command the test is to compare against
    description : string
    output_specs : list of strings
         The expected result of the command
    """

    def __init__(self, command, setup_commands=None, description="", output_specs=None):
        self.command = command
        self.description = description
        self.output_specs = [] if output_specs is None else output_specs
        self.setup_commands = [] if setup_commands is None else setup_commands


class OutputSpec:
    """
    Specification of the expected output of a test

    Attributes
    ----------
    type: scruf.compare.ComparisonTypes
    content: str
        The expected output content
    source: str
        The expected source for the line, one of "stdout", "stderr", or
        "returncode"
    """

    def __init__(
        self, comp_type=compare.ComparisonTypes.BASIC, content="", source="stdout"
    ):
        self.comp_type = comp_type
        self.content = content
        self.source = source

    def compare(self, result_content):
        return compare.get_comparer(self.comp_type)(self.content, result_content)
